open Ast

module NameMap = Map.Make(struct
  type t = string
  let compare x y = Pervasives.compare x y
end)

exception ReturnException of string NameMap.t * string NameMap.t NameMap.t

(* Main entry point: run a program *)

let run (vars, funcs) =
(* Put function declarations in a symbol table *)
  let func_decls = List.fold_left
  (fun funcs fdecl -> 
    if NameMap.mem fdecl.fname funcs then 
      raise (Failure ("function " ^ fdecl.fname ^ " is defined more than once!"));
      NameMap.add fdecl.fname fdecl funcs ) 
    NameMap.empty funcs
  in    
    let initall = List.fold_left
    (fun globals vdecl -> 
      let params = List.fold_left  
      (fun param_map param_decl -> NameMap.add param_decl.key param_decl.value param_map)
      NameMap.empty vdecl.varparams in
      
        if NameMap.mem vdecl.varname globals then 
          raise (Failure ("variable " ^ vdecl.varname ^ " is defined more than once!"));
          NameMap.add vdecl.varname params globals; 
        ) 
  in
  let rec call fdecl actuals globals =
    (* Evaluate an expression and return (value, updated environment) *)
    (**** This is the object eval section ****)
    let rec eval_object env = function
      Literal(i) -> i, env
      | Noexpr -> "Nothing", env (* must be non-zero for the for loop predicate *)
      | Str(str) -> (String.sub str 1 (( String.length str)-2)), env
      | Binop(e1, op, e2) ->
        let v1, env = eval_object env e1 in
          let v2, env = eval_object env e2 in
          let bool_to_str i = if i then "true" else "false" in
          (match op with
            Add -> string_of_int(int_of_string(v1) + int_of_string(v2))
            | Sub -> string_of_int(int_of_string(v1) - int_of_string(v2))
            | Mult -> string_of_int(int_of_string(v1) * int_of_string(v2))
            | Div -> string_of_int(int_of_string(v1) / int_of_string(v2))
            | Equal -> bool_to_str(int_of_string(v1) == int_of_string(v2))
            | Neq -> bool_to_str(int_of_string(v1) != int_of_string(v2))
            | Less -> bool_to_str(int_of_string(v1) < int_of_string(v2))
            | Leq -> bool_to_str(int_of_string(v1) <= int_of_string(v2))
            | Greater -> bool_to_str(int_of_string(v1) > int_of_string(v2))
            | Geq -> bool_to_str(int_of_string(v1) >= int_of_string(v2))
            | Combine -> String.concat "" (v1::(v2::[]))), env 
      | Access(var, param) ->
        let (locals, globals) = env in
          if NameMap.mem var locals then
            if NameMap.mem param (NameMap.find var locals) then
              NameMap.find param (NameMap.find var locals),(locals, globals)
            else raise (Failure ("undeclared identifier " ^ var))
          else if NameMap.mem var globals then
            if NameMap.mem param (NameMap.find var globals) then
              NameMap.find param (NameMap.find var globals),(locals, globals)
            else raise (Failure ("undeclared identifier " ^ var))
          else raise (Failure ("undeclared identifier " ^ var))
      | AssignToObject(var, param, e) ->
        let v, (locals, globals) = eval_object env e in
          if NameMap.mem var locals then
            v, (NameMap.add var (NameMap.add param v (NameMap.find var locals) ) locals, globals)
          else if NameMap.mem var globals then
            v, (locals, NameMap.add var (NameMap.add param v (NameMap.find var globals) ) globals) 
          else raise (Failure ("undeclared identifier " ^ var))
    in
    (* Evaluate an expression and return (value, updated environment) *)
    (**** This is the normal eval section ****)
    let rec eval env = function
        Id(var) ->
          let locals, globals = env in
            if NameMap.mem var locals then
             (NameMap.find var locals), env
            else if NameMap.mem var globals then
              (NameMap.find var globals), env 
            else raise (Failure ("undeclared identifier " ^ var))
        | Assign(var, param, e) ->
          let v, (locals, globals) = eval_object env e in
            if NameMap.mem var locals then
              NameMap.empty, (NameMap.add var (NameMap.add param v (NameMap.find var locals) ) locals, globals)
            else if NameMap.mem var globals then
              NameMap.empty, (locals, NameMap.add var (NameMap.add param v (NameMap.find var globals) ) globals) 
            else raise (Failure ("undeclared identifier " ^ var))
        | Call(f, actuals) ->
            let fdecl =
              try NameMap.find f func_decls
              with Not_found -> raise (Failure ("undefined function " ^ f))
          in
            let actuals, env = List.fold_left
                (fun (actuals, env) actual ->
            let v, env = eval env actual in v :: actuals, env)
              ([], env) actuals
            in
              let (locals, globals) = env in
              try
                let globals = call fdecl actuals globals in NameMap.empty, (locals, globals)
                  with ReturnException(v, globals) -> v, (locals, globals)
                in
                  (* Execute a statement and return an updated environment *)
                  let rec exec env = function
                    Block(stmts) -> List.fold_left exec env stmts
                    | Expr(e) -> let _, env = eval env e in env
                    | If(e, s1, s2) ->
                      let v, env = eval_object env e in
                      exec env (if String.compare "true" v==0 then s1 else s2)
                    | For(e1, e2, e3, s) ->
                      let _, env = eval_object env e1 in
                      let rec loop env =
                        let v, env = eval_object env e2 in
                        if String.compare "true" v==0 then
                          let _, env = eval_object (exec env s) e3 in
                          loop env
                        else
                          env
                      in loop env
                    | ToScreen(e) ->
                      let str, env = eval_object env e in
                        (match (str) with
                        | _ -> print_endline str ; flush stdout; env)
    in
      (* Enter the function: bind actual values to formal arguments *)
      let locals =
        try List.fold_left2
        (fun locals formal actual -> NameMap.add formal actual locals)
        NameMap.empty fdecl.formals actuals
        with Invalid_argument(_) ->
          raise (Failure ("wrong number of arguments passed to " ^ fdecl.fname))
      in
        (* Initilize local variables *)
        let locals = initall locals fdecl.locals in 
          (* Execute each statement; return updated global symbol table *)
          snd (List.fold_left exec (locals, globals) fdecl.body)
      in 
        (* add global variables to symbol table. *)
        let globals = initall NameMap.empty vars in 
  try
    call (NameMap.find "main" func_decls) [] globals
    with Not_found -> raise (Failure ("did not find the main() function"))